#! /usr/bin/env python
# -*- coding: utf_8 -*-
from pandac.PandaModules import loadPrcFileData 
loadPrcFileData("", "window-type none")         #серверу не требуется графическое окно
loadPrcFileData("", "audio-library-name null")  #и звуковая библиотека
from direct.showbase.ShowBase import ShowBase
from panda3d.core import *
from direct.task import Task

PORT = 9099     #порт, который будем использовать
CLIENTS = {}    #словарь, где будем хранить соединения

class MyServer(ShowBase): #наш класс приложения-сервера наследуется от ShowBase

    def __init__(self):
        ShowBase.__init__(self)

        self.cManager = QueuedConnectionManager()                       #менеджер, обслуживающий очередь соединений и данных
        self.cListener = QueuedConnectionListener(self.cManager, 0)     #класс, слушающий порт и принимающий входящие соединения
        self.cReader = QueuedConnectionReader(self.cManager, 0)         #класс, читающий данные с порта в собственный буффер
        self.cReader.setRawMode(True)                                   #мы устанавливаем raw режим (нам не требуются датаграммы)
        self.cWriter = ConnectionWriter(self.cManager,0)                #класс, отправляющий данные
        self.cWriter.setRawMode(True)                                   #аналогично - raw режим
        self.tcpSocket = self.cManager.openTCPServerRendezvous(PORT, 1) #открываем порт
        self.cListener.addConnection(self.tcpSocket)                    #собщаем менеджеру что нам нужно его прослушивать
        taskMgr.add(self.listenTask, "serverListenTask",-40)            #приём входящих соединений
        taskMgr.add(self.readTask, "serverReadTask", -39)               #приём данных
        taskMgr.add(self.disconnectTask, "serverDisconnectTask", -38)   #очистка отвалившихся соединений
        taskMgr.doMethodLater(5.0,self.broadcastTask,"serverBroadcastTask")#просто рассылаем сообщение "broadcast" раз в 5 секунд
        print "Server ready."
       
    def listenTask(self, task):                                         #зада обработки входящих соединеий
        if self.cListener.newConnectionAvailable():                     #если имеются необработанные соединения
            rendezvous = PointerToConnection()                          #создаём пустые указатели для заполнения на следующем шаге
            netAddress = NetAddress()                                   #здесь сказывается то, что ядро панды написано на сях, которые не столь гибкие (зато быстрые)
            newConnection = PointerToConnection()                       #так что вместо работы с результатом функции, мы передаём в функцию указатели которые эта функция должна заполнить
            if self.cListener.getNewConnection(rendezvous,netAddress,newConnection): #пытаемся заполнить указатели данными (получить соединеие)
                newConnection = newConnection.p()                       #получаем из указателя сам объект
                self.cReader.addConnection(newConnection)               #говорим ридеру, что нужно читать данные с полученного нами соединения
                CLIENTS[newConnection] = netAddress.getIpString()       #сохраняем соединение в словаре для дальнейшего использования
                print "Got a connection!"                               #^^^ вот оно! "mighty python" - в качестве ключей словаря мы 
            else:                                                       #^^^ используем непосредственно объект соединения, а в значение ставим IP
                print "getNewConnection returned false"
        return Task.cont

    def readTask(self, task):                                           #задача чтения данных
        while self.cReader.dataAvailable():                             #если в буффере что-то есть
            datagram = NetDatagram()                                    #создаём объект
            if self.cReader.getData(datagram):                          #и читаем в него данные из буффера
                msg = datagram.getMessage()                             #извлекаем из полученных данных собственно сообщение
                conn = datagram.getConnection()                         #и соединение которое это сообщение выдало
                self.cWriter.send(msg,conn)                             #в качестве теста отправляем полученное сообщение обратно - эхо
                #self.cWriter.send('HTTP/1.0 200 Ok\nContent-Type: text/html; charset=windows-1251\n\n<h1>Hellow World!</h1>',conn)
                #self.cManager.closeConnection(conn)
                print msg
        return Task.cont

    def disconnectTask(self, task):                                     #задача проверки отвалившихся соединений
        while self.cManager.resetConnectionAvailable():                 #если менеджер говорит, что есть сброшенные соединения
            connPointer = PointerToConnection()                         #создаём пустой указатель
            self.cManager.getResetConnection(connPointer)               #и получаем в него сброшенное соединение
            connection = connPointer.p()
            self.cReader.removeConnection(connection)                   #удаляем соединение из ридера
            if CLIENTS.has_key(connection):                             #и из списка соединений
                print CLIENTS[connection] + " reset connecton"
                del CLIENTS[connection]
        return task.cont

    def broadcastTask(self, task):                                      #рассылка сообщений по всему списку соединений
        for conn in CLIENTS:
            print "broadcast to " + CLIENTS[conn]                       
            self.cWriter.send("broadcast",conn)
        return task.again
    
if __name__ == "__main__":  #стандартное условие - если этот модуль является главным (запускаемым),
    app = MyServer()        #чтобы если мы захотим импотрировать класс, а не запускать файл 
    app.run()               #он не пытался создавать и стартовать экземпляр нашего приложения
